/* Copyright 1995 by Shay Cohen.
 * All rights reserved
 *
 * You may distribute this program freely, under some circumstances:
 * You may not remove my name from the source or documentation, mark
 * your changes if you make ones, don't change or remove this notice.
 * I am not responsible for any possible damage caused by this program,
 * and the responsiblty rests entirely with the user.
 *
 * You can reach me for bug reports, suggestions, comments and life
 * threats as follows:
 * InterNet - shay@aquanet.co.il.
 * FidoNet - 2:403/423 (Shay Cohen).
*/

#include <time.h>
#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include <fcntl.h>

#include "fido.h"
#include "config.h"
#include "misc.h"
#include "socket.h"

extern struct fidoaddress myfidoaddr,fakefidoaddr;
extern char buf[BUFSIZE];
extern int error;
extern char DefaultUsenetFolder[40];
extern char Emailaddr[80];
extern char mytimezone[4];
extern int postallowed;
extern char myNEWSGROUPSFILE[80];
extern char writepktfile[80];
extern char fromalonefolder[80];
extern char postpktfile[80];
extern char originline[70];
extern int lines_limit;
extern int bytes_limit;

int serverpost;
/*
   Purpose: Grab the new news messages ids from the server.

   Parameters:
   socket - socket to use.
   newsgroup - newsgroup name to scan.
   date - scan since date (yymmdd).
   time - scan since time (hhmmss).

   Returns: -1 if failed, else number of new messages.
*/

int nntp_newnews(socket, newsgroup, date, time)
int socket;
const char *newsgroup,*date,*time;
{
 FILE *f;
 int total=0,error;

 sprintf(buf,"NEWNEWS %s %s %s",newsgroup,date,time);
/* printf("-> %s\n",buf);*/

 error=SockPuts(socket,buf);
 if (error)
 {
  fprintf(stderr,ERROR_SOCKWRITE);
  return -1;
 }

 SockGets(socket,buf,BUFSIZE);
 /*printf("<- %s\n",buf);*/ buf[3] = '\0';
 if (strcmp("230",buf))
  {
   printf("Unexpected response from server in NEWNEWS command (%s)!\n",buf);
   return -1;
  }

 /* Open groups file, and put message ids there: */
 f=fopen(NEWNEWSFILE,"w");
 if (f==NULL)
 {
  fprintf(stderr,ERROR_FILEMSGID);
  return -1;
 }

 while ((error=SockGets(socket,buf,BUFSIZE)) || (error==error))
 {
  if (error)
  {
   fprintf(stderr,ERROR_SOCKREAD);
   fclose(f);
   return -1;
  }
  if (!strcmp(".",buf)) break;
  fprintf(f,"%s\n",buf);
  ++total;
 }

 /*if (opt.newnewsmsg)*/
   printf("Total new news in %s: %d\n",newsgroup,total);

 fclose(f);
 return total;
}


/* Purpose: get an article (by msg-id) size by using XHDR bytes */

int nntp_getarticlehdr(socket,hdr,messageid)
int socket;
char *hdr,*messageid;
{
 char buf[BUFSIZE],*s1;
 int size;

 sprintf(buf,"XHDR %s %s",hdr,messageid);

 error=SockPuts(socket,buf);
 if (error) {
   fprintf(stderr,ERROR_SOCKWRITE);
   return -1;
 }

 SockGets(socket,buf,BUFSIZE);
 buf[3] = '\0';
 if (strcmp("221",buf))
  {
   printf("Cannot get article size (ID: %s)!\n",messageid);
   return 0;
  }
 SockGets(socket,buf,BUFSIZE);
 s1=strtok(buf,">");

 s1=strtok(NULL,"");
 size=atoi(s1);

 SockGets(socket,buf,BUFSIZE);
 buf[1]=0;
 if (strcmp(".",buf)) {
		printf("Invalid terminate response for getting size!\n");
		return 0;
 }
 return size;
}

/*
   Purpose: Grab an article from the server, and output a fido message.

   Parameters:
   socket - socket to use.
   messageid - article's message-id.
   filename - fido message output filename.

   Returns: -1 if failed, else 0.
*/
int nntp_getarticle(socket,messageid,foldername,mytotal,ttotal)
int socket;
char *messageid, *foldername;
int mytotal,ttotal;
{
 char header[BUFSIZE],fromUserName[36],subject[72],
fromUserNameFull[BUFSIZE],NewsGroups[BUFSIZE],DateTime[20],DateTimeFull[40],MsgIdFull[120],OrgFull[120],*toUserName,
toUserName_[80];
 FILE *f;
 int total=0,error,article_lines,article_bytes;
 int articlestatus=HEADER,totalbyte=0;
 char nullchar=0;
 struct packedfidomsgheader fidoheader;
 

 if (recoverentryexists(messageid)) 
 {
  printf("Article was already retrieved before - %s",messageid);
  return 0;
 }

 /* Set the fido message header */
 memset(&fidoheader,0,sizeof(fidoheader));
 fidoheader.origNet = fakefidoaddr.net; 
 fidoheader.destNet = myfidoaddr.net;
 fidoheader.origNode = fakefidoaddr.node; 
 fidoheader.destNode = myfidoaddr.node;
 fidoheader.msgType = 0x0002;

 strcpy(OrgFull,"");

 article_lines=nntp_getarticlehdr(socket,"lines",messageid);
 article_bytes=nntp_getarticlehdr(socket,"bytes",messageid);
 if (article_lines>lines_limit) {
	printf("Article %s exceeds lines limit!\r\n",messageid);
	return 0;
 }
 if (article_bytes>bytes_limit) {
        printf("Article %s exceeds bytes limit!\r\n",messageid);
        return 0;
 }

 sprintf(buf,"ARTICLE %s",messageid);
  
 error=SockPuts(socket,buf);
 if (error) {
   fprintf(stderr,ERROR_SOCKWRITE);
   return -1;
 }
 SockGets(socket,buf,BUFSIZE);
/* printf("<- %s\n",buf);*/
 buf[3] = '\0';
 if (strcmp("220",buf))
  {
   printf("Cannot read article (ID: %s)!\n",messageid);
   return -1;
  }

 f=fopen(writepktfile,"a");
 fwrite(&fidoheader,1,sizeof(fidoheader),f);
 if (f==NULL) {
  fprintf(stderr,ERROR_FILEARTICLE);
 }

 while ((error=SockGets(socket,buf,BUFSIZE)) || (error==error))
 {
  if (error)
  {
   fprintf(stderr,ERROR_SOCKREAD);
   fclose(f);
   return -1;
  }
  if (!strcmp(".",buf)) break;
 
  totalbyte=totalbyte+strlen(buf);	

  if ((!strcmp("",buf)) && (articlestatus==HEADER))
  {
    articlestatus=BODY;
    toUserName=subjectexists(subject);
    if (toUserName==NULL) strcpy(toUserName_,"All");
          else strcpy(toUserName_,toUserName);
          
   /* printf("-# Reading body..\n");*/
    fprintf(f,"%s",DateTime); fwrite(&nullchar,1,sizeof(nullchar),f);
    fprintf(f,"%s",toUserName_); fwrite(&nullchar,1,sizeof(nullchar),f);
    fprintf(f,"%s",fromUserName); fwrite(&nullchar,1,sizeof(nullchar),f);
    fprintf(f,"%s",subject); fwrite(&nullchar,1,sizeof(nullchar),f);
    fprintf(f,"AREA:%s\r",foldername);
    
fprintf(f,"\01MSGID: %d:%d/%d %s\r%s\r%s\r%s\r%s\r%s\r\r",fakefidoaddr.zone,fakefidoaddr.net,fakefidoaddr.node,randomkludge(),MsgIdFull,fromUserNameFull,NewsGroups,DateTimeFull,OrgFull);
  }

  if (articlestatus==HEADER)
  {
   strcpy(header,articleheader(buf));
   if (!strcmp(header,"From"))
   {
    /*fprintf(f,"%s\r",buf);*/
    strcpy(fromUserNameFull,buf);
    strcpy(fromUserName,fromheader(buf));
    fromUserName[34]=0;
    printf("Article from: %s (%d of %d)\n",fromUserName,mytotal,ttotal);
    /*printf("-# setting from field from %s\n",fidoheader.fromUserName);*/
   }
   else
   if (!strcmp(header,"Subject"))
   {
    buf[72]='\0';
    strncpy(subject,&buf[9],strlen(buf)-8);
   /* printf("-# setting the field subject %s\n",fidoheader.subject);*/
   }
   else
   if (!strcmp(header,"Date"))
   {
    strcpy(DateTimeFull,buf);
    strcpy(DateTime,dateheader(buf));
    /*printf("-# setting the date field %s\n",fidoheader.DateTime);*/
   }
   if (!strcmp(header,"Message-ID"))
   { 
    strcpy(MsgIdFull,buf);
  /*  printf("Aritcle Msg-ID: %s\n",MsgIdFull);*/
   }
   else
   if (!strcmp(header,"Organization"))
   {
    strcpy(OrgFull,buf);
   }
   else
   if (!strcmp(header,"Newsgroups"))
   {
     strcpy(NewsGroups,buf);
     if (strlen(NewsGroups) > 77) 
              NewsGroups[77] = '\0';
   }
  }
   else 
  {
   if (buf[0]!='.')
    fprintf(f,"%s\r",buf);
     else fprintf(f,"%s\r",&buf[1]);
  } /* the article text */
  ++total;
 }
/* if (opt.articlemsg)
     printf("Read article (lines %d), wrote %s.\n",total,filename);*/
 
 fprintf(f,"\r--- %s\r * Origin: %s (%d:%d/%d)\r",MSG_TEAR,originline,myfidoaddr.zone,myfidoaddr.net,myfidoaddr.node);
 fprintf(f,"SEEN-BY: %d/%d\r",fakefidoaddr.net,fakefidoaddr.node);
 fprintf(f,"\1PATH: %d/%d\r",fakefidoaddr.net,fakefidoaddr.node);
 fwrite(&nullchar,1,sizeof(nullchar),f);

/* fseek(f,0,SEEK_SET);
 fwrite(&fidoheader,1,sizeof(fidoheader),f);*/
 fclose(f);

 if (totalbyte>60000)
 printf("Warning!\7 Article size: %dkb\n",totalbyte/1024);
 else
 printf("Article size: %dkb, %d line(s).\n",totalbyte/1024,article_lines);
 addrecoverentry(messageid);
 return 0;
}


int nntp_listgroups(sock)
int sock;
{
 char posting[20],groupname[BUFSIZE];
 FILE *f;
 int  temp;

 printf("Updating newsgroups file with ALL available newsgroups... Might take a while.\n");

 error=SockPuts(sock,"LIST");
 if (error)
 {
  fprintf(stderr,ERROR_SOCKWRITE);
  return -1;
 }
 SockGets(sock,buf,BUFSIZE); buf[3] = '\0';
 if (strcmp(buf,"215"))
 {
  printf("Unexpected response for a LIST command (%s).\n",buf);
  return -1;
 }

 f=fopen(myNEWSGROUPSFILE,"a+");
 if (f==NULL)
 {
  printf("Cannot open newsgroups file for output!\n");
  return -1;
 }
 temp=ftell(f);

 while ((error=SockGets(sock,buf,BUFSIZE)) || (error==error))
 {
  if (error) 
  {
   fprintf(stderr,ERROR_SOCKREAD);  
   fclose(f);
   return -1;
  }

  if (!strcmp(".",buf)) break;
 /* printf("%s\n",buf);*/
  
  strcat(buf,"\n");
  strcpy(groupname,strtok(buf," "));
  strtok(NULL," "); strtok(NULL," ");
  strcpy(posting,strtok(NULL,"\n"));
  strcpy(buf,groupname); strcpy(groupname,"*");
  strcat(groupname,buf);
  if ((((!temp) || (groupexists(f,groupname)==NULL))) 
           && (posting[0]!='=')) 
      addgrouptolist(f,groupname,posting,DefaultUsenetFolder,0);
 }
 fclose(f);
}




int nntp_getgroup(sock,folder,newsgroup,date,time)
int sock;
const char *folder,*newsgroup,*date,*time;
{
 int total,ttotal;
 char buf[BUFSIZE];
 FILE tf,*msgidfile;

 ttotal=nntp_newnews(sock,newsgroup,date,time);

 if (ttotal < 0)
 {
  fprintf(stderr,"?? Error in getting new news message-ids (%s)!\n",newsgroup);
  return -1;
 }
 else
 if (total == 0) return -1;

 total=0;
 
 msgidfile=fopen(NEWNEWSFILE,"r");

 if (msgidfile==NULL)
 {
  fprintf(stderr,"?? No new news file!");
  return -1;
 }

 while (fgets(buf,BUFSIZE,msgidfile) != NULL)
 {
  total++;
  if (nntp_getarticle(sock,buf,folder,total,ttotal) == -1)
  {
   fprintf(stderr,"?? Cannot get new article to folder %s\n",folder);
   total--;
  }
 }
 printf("Total articles %d grabbed from %s.\n",total,newsgroup);
 
 fclose(msgidfile);
}


int nntp_scangroups(sock)
int sock;
{
 int curpos,total=0;
 char newsgroup[BUFSIZE],ttime[7],date[7],posting[4],directory[40];
 FILE *f;
 time_t thetime;

 printf("Creating fido packet %s and scanning groups...\n",writepktfile);

 f=fopen(myNEWSGROUPSFILE,"r+");
 if (f==NULL)
 {
  fprintf(stderr,"Cannot open %s file! Please set one.\n",myNEWSGROUPSFILE);
  return -1; 
 }
 while (1)
 {
  curpos=ftell(f);
  if (fgets(buf,BUFSIZE,f) == NULL) break;
  
  if ((index(buf,':') != NULL) && (buf[0]!='#'))
                            {

  strcpy(newsgroup,strtok(buf,":"));
  strcpy(date,strtok(NULL,":"));
  strcpy(ttime,strtok(NULL,":"));
  strcpy(posting,strtok(NULL,":"));
  strcpy(directory,strtok(NULL,":"));
 
  directory[strlen(directory)-1] = '\0';  

  if (newsgroup[0]=='*')
  {
/*   printf("%s is inactive, skipped.\n",newsgroup);*/
  }
  else
  {
  printf("Scanning %s (%s/%s,->%s)\n",newsgroup,date,ttime,directory);
  
  thetime=time(&thetime);

  if (nntp_getgroup(sock,directory,newsgroup,date,ttime)==0)
   {
    fseek(f,curpos,SEEK_SET);
    addgrouptolist(f,newsgroup,posting,directory,thetime);
    ++total;
   }
  }
                            }
 }
 printf("Total groups scanned %d\n",total);
 fclose(f);
}

int nextstr(f,s)
FILE *f;
char *s;
{
 char ch=1;
 while (ch)
 {
  fread(&ch,1,1,f);
  *s++=ch;
 }
}

int getnextline(f,s)
FILE *f;
char *s;
{
 char ch=1;
 int tempftell;
 while ((ch) && (ch!='\r')/* && (ch!='\n')*/)
 {
  fread(&ch,1,1,f);

  if ((ch!='\r') && (ch!='\n')) *s++=ch;
 }
/* *s--;*/
 *s='\0';
/* if (ch!='\0')
 {
 tempftell=ftell(f);
 fread(&ch,1,1,f);
 if ((ch!='\r') && (ch!='\n')) {
          fseek(f,tempftell,SEEK_SET);
       }
 }*/
 if (ch=='\0') return ENDOFMSG; 
  else return NEWLINE;
}

int postmail(f)
FILE *f;
{
 int filepos,getresult;
 FILE myoutf,*outf=&myoutf;
 struct packedfidomsgheader msgheader;
 char toName[36],fromName[36],subject[72],date[20],line[255]="\1",emailaddr[80],syscall[200];
 char temp[12],temp2[12],temp3[12];
 
 filepos=ftell(f);
 fread(&msgheader,1,sizeof(msgheader),f);
 if (msgheader.msgType!=0x02) return 2;
 
 nextstr(f,date);
 nextstr(f,toName);
 nextstr(f,fromName);
 nextstr(f,subject);



 if (strcmp(toName,"EMAIL"))
 {
  fseek(f,filepos,SEEK_SET);
  return -2;
 }

 outf=fopen("/tmp/temp.msg","w");
 if (outf==NULL)
 {
  fprintf(stderr,"Error posting e-mail.\n");
  while (getnextline(f,line)!=ENDOFMSG);
  return -1;
 }

 fprintf(outf,"From: %s <%s>\r\n",fromName,Emailaddr);
 fprintf(outf,"Subject: %s\r\n\r\n",subject);

 line[0]='\1';
 temp[4]='\0';
 while ((line[0]=='\1') || (!strcmp(strncpy(temp,line,4),"AREA")) )
 {
  getresult=getnextline(f,line);
  if (getresult==ENDOFMSG)
  {
   fprintf(stderr,"Cannot find mail address!");
   return -1;
  }
 }

 strcpy(emailaddr,line);
 if (getresult==ENDOFMSG) return 0;
 getresult=getnextline(f,line);
 if (getresult==ENDOFMSG) return 0;

 while ((getresult=getnextline(f,line))!=ENDOFMSG)
 {
  strncpy(temp,line,3); temp[3]='\0';
  strncpy(temp2,line,10); temp2[10]='\0';
  strncpy(temp3,line,8); temp3[8]='\0';
  if ((line[0]!='\1') && (strcmp(temp,"---")) &&
            (strcmp(temp2," * Origin:")) && 
                    (strcmp(temp3,"SEEN-BY:")))
  fprintf(outf,"%s\r\n",line);
 }
 fclose(outf);
 sprintf(syscall,"sendmail %s </tmp/temp.msg",emailaddr);
 fprintf(stderr,"Executing mail...\n");
 system(syscall);
 fprintf(stderr,"Posted e-mail.\n");
 return 0;
}

int postmessage(sock,f)
int sock;
FILE *f;
{
 struct packedfidomsgheader msgheader;
 char toName[36],fromName[36],subject[72],date[20],folder[255]="\1",
      line[255],temp[12],temp2[12],temp3[12],temp4[12],*group/*,mydate[35]*/;
 int getresult;


 fread(&msgheader,1,sizeof(msgheader),f);
 if (msgheader.msgType != 0x0002) return 2;

 error=SockPuts(sock,"POST");
 if (error)
 {
   fprintf(stderr,ERROR_SOCKREAD);
   while (getnextline(f,folder)!=ENDOFMSG);
   return -1;
 }
 SockGets(sock,buf,BUFSIZE);
 buf[3]='\0';
 if (strcmp(buf,"340"))
 {
   printf("Article cannot be posted, invalid response!\n");
   while (getnextline(f,folder)!=ENDOFMSG);
   return -1;
 }

 nextstr(f,date);
 nextstr(f,toName);
 nextstr(f,fromName);
 nextstr(f,subject);

 addsubject(fromName,subject);
 
 while (folder[0] == '\1')
 {
   getresult=getnextline(f,folder);
   if (getresult==ENDOFMSG)
   {
     fprintf(stderr,ERROR_NOAREA);
     return -1;
   }
 }

 if (strstr(folder,"AREA:")==NULL)
 {
  fprintf(stderr,ERROR_NOAREA);
  while (getnextline(f,folder)!=ENDOFMSG);
  return -1;
 }

 strncpy(folder,&folder[5],strlen(folder)-5);
 folder[strlen(folder)-5] = '\0';
 group=foldertonewsgroup(folder);
 if (!strcmp(folder,fromalonefolder)) /* Let me alone folder? */ 
          {
             group=&folder[0]; /* Put a dummy, so group won't be NULL */
             postallowed=ALLOWED;
          }
 if (group==NULL)
 {
  fprintf(stderr,ERROR_NOGROUP);
  while (getnextline(f,folder)!=ENDOFMSG);
  return -1;
 }
 if (!postallowed)
 {
  fprintf(stderr,ERROR_NOTALLOWED); 
  while (getnextline(f,folder)!=ENDOFMSG);
  return -1;
 }

 /* removed the from section, that way we can have whatever from header
		we want */
 /*sprintf(buf,"From: %s (%s)",Emailaddr,fromName);
 SockPuts(sock,buf);
  */

 sprintf(buf,"Subject: %s",subject);
 SockPuts(sock,buf);

 strncpy(temp,date,2); temp[2]='\0';
 strncpy(temp2,&date[3],3); temp2[3]='\0';
 strncpy(temp3,&date[7],2); temp3[2]='\0';
 strncpy(temp4,&date[11],8); temp4[8]='\0';
/* sprintf(mydate,"Date: Day, %s-%s-%s %s 
%s",temp,temp2,temp3,temp4,mytimezone);
 SockPuts(sock,mydate);*/

 if (strcmp(folder,fromalonefolder))
 {
  sprintf(buf,"Newsgroups: %s\n",group);
  SockPuts(sock,buf);
 }

 while ((getresult=getnextline(f,line)) || (getresult==getresult))
 {
  if (getresult==ENDOFMSG) break;
  if (line[0]=='.') {
   strcpy(buf,".");
   strcat(buf,line);
   strcpy(line,buf);
  } 
  strncpy(temp,line,3); temp[3]='\0';
  strncpy(temp2,line,10); temp2[10]='\0';
  strncpy(temp3,line,8); temp3[8]='\0';
  if ((line[0]!='\1') && (strcmp(temp,"---")) &&
            (strcmp(temp2," * Origin:")) && 
                    (strcmp(temp3,"SEEN-BY:")))
                             SockPuts(sock,line);
 }
 SockPuts(sock,MSG_TEAR);
 SockPuts(sock,".");
 SockGets(sock,buf,BUFSIZE);
 buf[3]='\0';
 if (strcmp(buf,"240"))
 {
  printf("Article wasn't posted ok! (%s)\n",&buf[4]);
  return -1;
 }
 printf("Article from %s in folder %s posted.\n",fromName,folder);
 return 0;
}

int postnext(sock,f)
int sock;
FILE *f;
{
 int result;
 result=postmail(f);
 if (result==-2)
 return postmessage(sock,f); 
 else return result;
}

int postpkt(sock,filename)
int sock;
char *filename;
{
 FILE *f;
 struct fidopacketheader pktheader;

 if (serverpost!=ALLOWED)
 {
  printf("Cannot post messages on this server!\n");
  return -1;
 } 
 printf("Posting %s...\n",filename);

 f=fopen(filename,"r");
 if (f==NULL) 
 {
  fprintf(stderr,ERROR_PKTOPEN);
  return -1;
 }
 fread(&pktheader,1,sizeof(pktheader),f);
 while (postnext(sock,f) != 2);
 fclose(f);
 printf("Removing %s (uploaded packet)...\n",filename);
 unlink(filename);
}

int nntp_greet(sock)
int sock;
{
 SockGets(sock,buf,BUFSIZE);
 printf("%s",buf);
 buf[3]='\0';
 if (!strcmp(buf,"200"))
 {
  printf("Connected to the server, posting allowed\n");
  serverpost=ALLOWED;
  return 0;
 }
 else
 if (!strcmp(buf,"201"))
 {
  printf("Connected to the server, posting is not allowed\n");
  serverpost=NOTALLOWED;
  return 0;
 }
 else
 if (!strcmp(buf,"502"))
 {
  printf("Connected to a server which doesn't like you.\n");
  return -1;
 }
 else
 {
  printf("Connected to the server, unexpected response, closing connection.\n");
  return -1;
 }
}
